'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var Ui = require('kenga/utils');
var Anchors = require('kenga-containers/anchors-pane');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var Ui__default = /*#__PURE__*/_interopDefaultLegacy(Ui);
var Anchors__default = /*#__PURE__*/_interopDefaultLegacy(Anchors);

function leftInSpaceOf(elem, parent) {
  var left = 0;
  var curr = elem.offsetParent; // This intentionally excludes body which has a null offsetParent.

  while (curr && curr !== parent && curr.offsetParent) {
    left -= curr.scrollLeft;
    curr = curr.parentNode;
  }

  while (elem && elem !== parent) {
    left += elem.offsetLeft;
    elem = elem.offsetParent;
  }

  return left;
}

function topInSpaceOf(elem, parent) {
  var top = 0;
  var curr = elem.offsetParent; // This intentionally excludes body which has a null offsetParent.

  while (curr && curr !== parent && curr.offsetParent) {
    top -= curr.scrollTop;
    curr = curr.parentNode;
  }

  while (elem && elem !== parent) {
    top += elem.offsetTop;
    elem = elem.offsetParent;
  }

  return top;
}

function mouseDrag(resizer, onBegin, onChange, onEnd) {
  var mouseDownAt = null;
  var prevState = null;
  var onMouseUp = null;
  var onMouseMove = null;
  var onDragStart = Ui__default["default"].on(resizer, Ui__default["default"].Events.DRAGSTART, function (event) {
    if (event.button === 0) {
      event.preventDefault();
      event.stopPropagation();
    }
  });
  var onMouseDown = Ui__default["default"].on(resizer, Ui__default["default"].Events.MOUSEDOWN, function (event) {
    if (event.button === 0) {
      event.stopPropagation();
      mouseDownAt = {
        x: event.clientX,
        y: event.clientY
      };
      prevState = onBegin(event);

      if (!onMouseUp) {
        onMouseUp = Ui__default["default"].on(document, Ui__default["default"].Events.MOUSEUP, function (event) {
          event.stopPropagation();

          if (onMouseUp) {
            onMouseUp.removeHandler();
            onMouseUp = null;
          }

          if (onMouseMove) {
            onMouseMove.removeHandler();
            onMouseMove = null;
          }

          var mouseDiff = {
            x: event.clientX - mouseDownAt.x,
            y: event.clientY - mouseDownAt.y
          };
          onEnd(prevState, mouseDiff, event);
        });
      }

      if (!onMouseMove) {
        onMouseMove = Ui__default["default"].on(document, Ui__default["default"].Events.MOUSEMOVE, function (event) {
          event.stopPropagation();
          event.preventDefault();
          var mouseDiff = {
            x: event.clientX - mouseDownAt.x,
            y: event.clientY - mouseDownAt.y
          };
          onChange(prevState, mouseDiff, event);
        });
      }
    }
  });
  return {
    removeHandler: function removeHandler() {
      onDragStart.removeHandler();
      onMouseDown.removeHandler();
    }
  };
}

function sizeLocationSnapshot(subject) {
  return {
    left: subject.left,
    top: subject.top,
    width: subject.width,
    height: subject.height,
    anchors: {
      left: subject.element.style.left,
      width: subject.element.style.width,
      right: subject.element.style.right,
      top: subject.element.style.top,
      height: subject.element.style.height,
      bottom: subject.element.style.bottom
    }
  };
}

function applySizeLocationSnapshot(snapshot, subject) {
  subject.element.style.left = snapshot.anchors.left;
  subject.element.style.width = snapshot.anchors.width;
  subject.element.style.right = snapshot.anchors.right;
  subject.element.style.top = snapshot.anchors.top;
  subject.element.style.height = snapshot.anchors.height;
  subject.element.style.bottom = snapshot.anchors.bottom;
}

function resizeDecor(surface, subject, onEnd) {
  return {
    lt: function () {
      var lt = document.createElement('div');
      lt.className = 'p-winnie-decoration p-winnie-decoration-left-top';
      mouseDrag(lt, function () {
        return sizeLocationSnapshot(subject);
      }, function (prevState, diff) {
        subject.left = prevState.left + diff.x;
        subject.top = prevState.top + diff.y;
        subject.width = prevState.width - diff.x;
        subject.height = prevState.height - diff.y;
        lt.neightbours.forEach(function (d) {
          d.stick();
        });
      }, onEnd);

      lt.stick = function () {
        lt.style.left = "".concat(leftInSpaceOf(subject.element, surface.element), "px");
        lt.style.top = "".concat(topInSpaceOf(subject.element, surface.element), "px");
        lt.style.display = subject.element.offsetParent != null ? '' : 'none';
      };

      return lt;
    }(),
    lm: function () {
      var lm = document.createElement('div');
      lm.className = 'p-winnie-decoration p-winnie-decoration-left-middle';
      mouseDrag(lm, function () {
        return sizeLocationSnapshot(subject);
      }, function (prevState, diff) {
        subject.left = prevState.left + diff.x;
        subject.width = prevState.width - diff.x;
        lm.neightbours.forEach(function (d) {
          d.stick();
        });
      }, onEnd);

      lm.stick = function () {
        lm.style.left = "".concat(leftInSpaceOf(subject.element, surface.element), "px");
        lm.style.top = "".concat(topInSpaceOf(subject.element, surface.element) + subject.element.offsetHeight / 2, "px");
        lm.style.display = subject.element.offsetParent != null ? '' : 'none';
      };

      return lm;
    }(),
    lb: function () {
      var lb = document.createElement('div');
      lb.className = 'p-winnie-decoration p-winnie-decoration-left-bottom';
      mouseDrag(lb, function () {
        return sizeLocationSnapshot(subject);
      }, function (prevState, diff) {
        subject.left = prevState.left + diff.x;
        subject.width = prevState.width - diff.x;
        subject.height = prevState.height + diff.y;
        lb.neightbours.forEach(function (d) {
          d.stick();
        });
      }, onEnd);

      lb.stick = function () {
        lb.style.left = "".concat(leftInSpaceOf(subject.element, surface.element), "px");
        lb.style.top = "".concat(topInSpaceOf(subject.element, surface.element) + subject.element.offsetHeight, "px");
        lb.style.display = subject.element.offsetParent != null ? '' : 'none';
      };

      return lb;
    }(),
    mt: function () {
      var mt = document.createElement('div');
      mt.className = 'p-winnie-decoration p-winnie-decoration-middle-top';
      mouseDrag(mt, function () {
        return sizeLocationSnapshot(subject);
      }, function (prevState, diff) {
        subject.top = prevState.top + diff.y;
        subject.height = prevState.height - diff.y;
        mt.neightbours.forEach(function (d) {
          d.stick();
        });
      }, onEnd);

      mt.stick = function () {
        mt.style.left = "".concat(leftInSpaceOf(subject.element, surface.element) + subject.element.offsetWidth / 2, "px");
        mt.style.top = "".concat(topInSpaceOf(subject.element, surface.element), "px");
        mt.style.display = subject.element.offsetParent != null ? '' : 'none';
      };

      return mt;
    }(),
    mb: function () {
      var mb = document.createElement('div');
      mb.className = 'p-winnie-decoration p-winnie-decoration-middle-bottom';
      mouseDrag(mb, function () {
        return sizeLocationSnapshot(subject);
      }, function (prevState, diff) {
        subject.height = prevState.height + diff.y;
        mb.neightbours.forEach(function (d) {
          d.stick();
        });
      }, onEnd);

      mb.stick = function () {
        mb.style.left = "".concat(leftInSpaceOf(subject.element, surface.element) + subject.element.offsetWidth / 2, "px");
        mb.style.top = "".concat(topInSpaceOf(subject.element, surface.element) + subject.element.offsetHeight, "px");
        mb.style.display = subject.element.offsetParent != null ? '' : 'none';
      };

      return mb;
    }(),
    rt: function () {
      var rt = document.createElement('div');
      rt.className = 'p-winnie-decoration p-winnie-decoration-right-top';
      mouseDrag(rt, function () {
        return sizeLocationSnapshot(subject);
      }, function (prevState, diff) {
        subject.top = prevState.top + diff.y;
        subject.width = prevState.width + diff.x;
        subject.height = prevState.height - diff.y;
        rt.neightbours.forEach(function (d) {
          d.stick();
        });
      }, onEnd);

      rt.stick = function () {
        rt.style.left = "".concat(leftInSpaceOf(subject.element, surface.element) + subject.element.offsetWidth, "px");
        rt.style.top = "".concat(topInSpaceOf(subject.element, surface.element), "px");
        rt.style.display = subject.element.offsetParent != null ? '' : 'none';
      };

      return rt;
    }(),
    rm: function () {
      var rm = document.createElement('div');
      rm.className = 'p-winnie-decoration p-winnie-decoration-right-middle';
      mouseDrag(rm, function () {
        return sizeLocationSnapshot(subject);
      }, function (prevState, diff) {
        subject.width = prevState.width + diff.x;
        rm.neightbours.forEach(function (d) {
          d.stick();
        });
      }, onEnd);

      rm.stick = function () {
        rm.style.left = "".concat(leftInSpaceOf(subject.element, surface.element) + subject.element.offsetWidth, "px");
        rm.style.top = "".concat(topInSpaceOf(subject.element, surface.element) + subject.element.offsetHeight / 2, "px");
        rm.style.display = subject.element.offsetParent != null ? '' : 'none';
      };

      return rm;
    }(),
    rb: function () {
      var rb = document.createElement('div');
      rb.className = 'p-winnie-decoration p-winnie-decoration-right-bottom';
      mouseDrag(rb, function () {
        return sizeLocationSnapshot(subject);
      }, function (prevState, diff) {
        subject.width = prevState.width + diff.x;
        subject.height = prevState.height + diff.y;
        rb.neightbours.forEach(function (d) {
          d.stick();
        });
      }, onEnd);

      rb.stick = function () {
        rb.style.left = "".concat(leftInSpaceOf(subject.element, surface.element) + subject.element.offsetWidth, "px");
        rb.style.top = "".concat(topInSpaceOf(subject.element, surface.element) + subject.element.offsetHeight, "px");
        rb.style.display = subject.element.offsetParent != null ? '' : 'none';
      };

      return rb;
    }()
  };
}

function startItemsMove(model, pickedWidget) {
  var pickedItem = pickedWidget['winnie.wrapper'];
  return function () {
    return model.layout.explorer.isSelected(pickedItem) ? model.layout.explorer.selected : [pickedItem];
  }().filter(function (item) {
    return item.delegate.attached && item.delegate.parent instanceof Anchors__default["default"];
  }).map(function (item) {
    return {
      item: item,
      startSnapshot: sizeLocationSnapshot(item.delegate)
    };
  });
}

function proceedItemsMove(model, items, diff, snap) {
  function snapX(v) {
    return snap ? Math.round(v / model.settings.grid.x) * model.settings.grid.x : v;
  }

  function snapY(v) {
    return snap ? Math.round(v / model.settings.grid.y) * model.settings.grid.y : v;
  }

  items.forEach(function (moved) {
    if (diff.x !== 0) {
      var left = moved.startSnapshot.left + diff.x;
      moved.item.delegate.left = snapX(left);
    }

    if (diff.y !== 0) {
      var top = moved.startSnapshot.top + diff.y;
      moved.item.delegate.top = snapY(top);
    }

    model.stickDecors();
  });
}

function endItemsMove(model, items) {
  if (items.length > 0) {
    items.forEach(function (moved) {
      moved.endSnapshot = sizeLocationSnapshot(moved.item.delegate);
    });
    model.edit({
      name: items.length === 1 ? "Widget '".concat(items[0].item.name, "' moved") : "Move of (".concat(items.length, ") widgets"),
      redo: function redo() {
        items.forEach(function (moved) {
          applySizeLocationSnapshot(moved.endSnapshot, moved.item.delegate);
        });
        model.stickDecors();
      },
      undo: function undo() {
        items.forEach(function (moved) {
          applySizeLocationSnapshot(moved.startSnapshot, moved.item.delegate);
        });
        model.stickDecors();
      }
    });
  }
}

exports.applySizeLocationSnapshot = applySizeLocationSnapshot;
exports.endItemsMove = endItemsMove;
exports.mouseDrag = mouseDrag;
exports.proceedItemsMove = proceedItemsMove;
exports.resizeDecor = resizeDecor;
exports.sizeLocationSnapshot = sizeLocationSnapshot;
exports.startItemsMove = startItemsMove;
