'use strict';

var _classCallCheck = require('@babel/runtime/helpers/classCallCheck');
var _createClass = require('@babel/runtime/helpers/createClass');
var Logger = require('septima-utils/logger');
var Ui = require('kenga/utils');
var Bound = require('kenga/bound');
var Widget = require('kenga/widget');
var Container = require('kenga/container');
var Label = require('kenga-labels/label');
var Box = require('kenga-containers/box-pane');
var Toolbar = require('kenga-containers/tool-bar');
var FlowPane = require('kenga-containers/flow-pane');
var ScrollPane = require('kenga-containers/scroll-pane');
var GridPane = require('kenga-containers/grid-pane');
var Anchors = require('kenga-containers/anchors-pane');
var TabbedPane = require('kenga-containers/tabbed-pane');
var HolyGrailPane = require('kenga-containers/holy-grail-pane');
var CardsPane = require('kenga-containers/card-pane');
var Menu = require('kenga-menu/menu');
var MenuItem = require('kenga-menu/menu-item');
var MenuElement = require('kenga-menu/menu-element');
var DataGrid = require('kenga-grid/grid');
var ColumnNode = require('kenga-grid/columns/column-node');
var textNumberField = require('./text-number-field.js');
var ground = require('./ground.js');
var i18n = require('./i18n.js');
var winnieWidget = require('./winnie-widget.js');
var rewidth = require('./rewidth.js');
var render = require('./properties/render.js');
require('./properties/edit.js');
var shortcuts = require('./shortcuts.js');
var rectSelection = require('./rect-selection.js');
var locationSize = require('./location-size.js');
var modelToCode = require('./serial/model-to-code.js');
var modelToJson = require('./serial/model-to-json.js');
var clipboard = require('./clipboard.js');
var widgets = require('./model/widgets.js');
var classes = require('./model/classes.js');
var properties = require('./model/properties.js');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var _classCallCheck__default = /*#__PURE__*/_interopDefaultLegacy(_classCallCheck);
var _createClass__default = /*#__PURE__*/_interopDefaultLegacy(_createClass);
var Logger__default = /*#__PURE__*/_interopDefaultLegacy(Logger);
var Ui__default = /*#__PURE__*/_interopDefaultLegacy(Ui);
var Bound__default = /*#__PURE__*/_interopDefaultLegacy(Bound);
var Widget__default = /*#__PURE__*/_interopDefaultLegacy(Widget);
var Container__default = /*#__PURE__*/_interopDefaultLegacy(Container);
var Label__default = /*#__PURE__*/_interopDefaultLegacy(Label);
var Box__default = /*#__PURE__*/_interopDefaultLegacy(Box);
var Toolbar__default = /*#__PURE__*/_interopDefaultLegacy(Toolbar);
var FlowPane__default = /*#__PURE__*/_interopDefaultLegacy(FlowPane);
var ScrollPane__default = /*#__PURE__*/_interopDefaultLegacy(ScrollPane);
var GridPane__default = /*#__PURE__*/_interopDefaultLegacy(GridPane);
var Anchors__default = /*#__PURE__*/_interopDefaultLegacy(Anchors);
var TabbedPane__default = /*#__PURE__*/_interopDefaultLegacy(TabbedPane);
var HolyGrailPane__default = /*#__PURE__*/_interopDefaultLegacy(HolyGrailPane);
var CardsPane__default = /*#__PURE__*/_interopDefaultLegacy(CardsPane);
var Menu__default = /*#__PURE__*/_interopDefaultLegacy(Menu);
var MenuItem__default = /*#__PURE__*/_interopDefaultLegacy(MenuItem);
var MenuElement__default = /*#__PURE__*/_interopDefaultLegacy(MenuElement);
var DataGrid__default = /*#__PURE__*/_interopDefaultLegacy(DataGrid);
var ColumnNode__default = /*#__PURE__*/_interopDefaultLegacy(ColumnNode);

function decapitalize(name) {
  return name.substring(0, 1).toLowerCase() + name.substring(1);
}

var Winnie = /*#__PURE__*/function () {
  function Winnie() {
    _classCallCheck__default["default"](this, Winnie);

    var enabled = [];

    function checkEnabled() {
      Ui__default["default"].later(function () {
        enabled.forEach(function (item) {
          item();
        });
      });
    }

    this.checkEnabled = checkEnabled;
    var self = this;
    this.settings = {
      grid: {
        x: 10,
        y: 10,
        snap: true
      },
      undoDepth: 1024 * 1024
    };
    this.paletteDrag = null;
    this.widgets = new Map(); // name -> widget

    this.forest = []; // root widgets

    this.layout = ground();
    this.edits = [];
    this.editsCursor = 0;
    this.decors = [];
    this.layout.leftSizer.element.classList.add('p-winnie-left-sizer');
    this.layout.rightSizer.element.classList.add('p-winnie-right-sizer');
    this.layout.propertiesHeader.element.classList.add('p-winnie-properties-header');
    this.layout.palette.element.classList.add('p-winnie-palette');
    this.layout.explorer.element.classList.add('p-winnie-explorer');
    this.layout.paletteExplorerSplit.element.classList.add('p-winnie-palette-explorer');
    this.layout.ground.element.classList.add('p-winnie-ground');
    this.layout.view.element.classList.add('p-winnie-view');
    this.layout.widgets.element.classList.add('p-winnie-widgets');
    this.hints = document.createElement('div');
    this.hints.className = 'p-winnie-hints';
    this.hints.innerHTML = '' + '<ul>' + "<li class=\"p-winnie-hint p-winnie-add-hint\">".concat(i18n['winnie.add.hint'], "</li>") + // `<li class="p-winnie-hint p-winnie-edit-hint">${i18n['winnie.edit.hint']}</li>` +
    '</ul>';
    this.layout.widgets.element.appendChild(this.hints);
    Ui__default["default"].on(this.layout.ground.element, Ui__default["default"].Events.KEYDOWN, function (event) {
      shortcuts.winnieKeyDown(self, event);
    });
    Ui__default["default"].on(this.layout.ground.element, 'paste', function (event) {
      if (!self.layout.explorer.activeEditor && !self.layout.properties.activeEditor) {
        event.stopPropagation();
        event.preventDefault();
        var firstContainer = null;
        var addLog = self.paste(event.clipboardData.getData('text/plain'));
        addLog.forEach(function (item) {
          if (!firstContainer && item.subject.delegate instanceof Container__default["default"] && !item.subject.delegate.parent) {
            firstContainer = item.subject;
          }
        });

        if (!self.visualRootPresent() && firstContainer) {
          self.acceptVisualRoot(firstContainer);
          self.layout.explorer.goTo(firstContainer, true);
          self.centerSurface();
        }
      }
    });
    Ui__default["default"].on(this.layout.widgets.element, Ui__default["default"].Events.KEYDOWN, function (event) {
      shortcuts.surfaceKeyDown(self, event);
    });
    Ui__default["default"].on(this.layout.widgets.element, Ui__default["default"].Events.CLICK, function (event) {
      if (event.target.className.includes('p-tab-caption-close-tool') || event.target.parentElement.className.includes('p-tab-caption-close-tool')) {
        event.stopPropagation();
      }
    }, true);
    locationSize.mouseDrag(this.layout.view.element, function (event) {
      return rectSelection.startRectSelection(self.layout.view.element, event);
    }, rectSelection.proceedRectSelection, function (start, diff, event) {
      rectSelection.endRectSelection(self, start, diff, event);
    });
    this.layout.explorer.data = this.forest;
    this.layout.explorer.parentField = 'parent';
    this.layout.explorer.childrenField = 'children';

    this.layout.explorer.onSelect = function (evt) {
      self.lastSelected = evt.item;
      checkEnabled();
    };

    this.layout.widgetColumn.onRender = function (item, viewCell) {
      viewCell.innerHTML = "".concat(item.name, " [").concat(item.source.name, "]");
      viewCell.title = "'".concat(item.source.name, "' from '").concat(item.source.from, "'");

      if (item.delegate.element === self.visualRootElement()) {
        viewCell.classList.add('p-winnie-visual-root');
      }
    };

    function isParent(parent, child) {
      while (child.parent && child.parent !== parent) {
        child = child.parent;
      }

      return !!child.parent;
    }

    function move(w, dest, addAt, parentUR) {
      if (!dest || dest.delegate instanceof Container__default["default"] && w.delegate instanceof Widget__default["default"] || dest.delegate instanceof DataGrid__default["default"] && w.delegate instanceof ColumnNode__default["default"] || dest.delegate instanceof ColumnNode__default["default"] && w.delegate instanceof ColumnNode__default["default"]) {
        if (w !== dest) {
          if (!dest || !isParent(w, dest)) {
            var source = w.parent;

            if (!dest || !dest.full || dest === source) {
              var removeAt = source ? source.indexOf(w) : self.forest.indexOf(w);

              if (source === dest && addAt > removeAt) {
                addAt--;
              }

              if (source !== dest || removeAt !== addAt) {
                var oldLocationSize = w.delegate instanceof ColumnNode__default["default"] ? {} : locationSize.sizeLocationSnapshot(w.delegate);

                if (dest) {
                  if (source !== dest) {
                    if (dest.delegate instanceof TabbedPane__default["default"] || dest.delegate instanceof CardsPane__default["default"] || dest.delegate instanceof GridPane__default["default"] || dest.delegate instanceof HolyGrailPane__default["default"] && addAt === 2) {
                      w.delegate.element.style.left = '';
                      w.delegate.element.style.width = '';
                      w.delegate.element.style.right = '';
                      w.delegate.element.style.top = '';
                      w.delegate.element.style.height = '';
                      w.delegate.element.style.bottom = '';
                    } else if (dest.delegate instanceof Box__default["default"]) {
                      w.delegate.element.style.left = '';
                      w.delegate.element.style.top = '';
                      w.delegate.element.style.right = '';
                      w.delegate.element.style.bottom = '';

                      if (dest.delegate.orientation === Ui__default["default"].Orientation.VERTICAL) {
                        w.delegate.element.style.width = '';
                      } else {
                        w.delegate.element.style.height = '';
                      }
                    } else if (dest.delegate instanceof HolyGrailPane__default["default"]) {
                      if (addAt === 0 || addAt === 4) {
                        w.delegate.element.style.left = '';
                        w.delegate.element.style.width = '';
                        w.delegate.element.style.right = '';
                        w.delegate.element.style.top = '';
                        w.delegate.element.style.bottom = '';
                      } else if (addAt === 1 || addAt === 3) {
                        w.delegate.element.style.left = '';
                        w.delegate.element.style.height = '';
                        w.delegate.element.style.right = '';
                        w.delegate.element.style.top = '';
                        w.delegate.element.style.bottom = '';
                      }
                    }
                  }
                }

                var newLocationSize = w.delegate instanceof ColumnNode__default["default"] ? {} : locationSize.sizeLocationSnapshot(w.delegate);
                var ur = {
                  name: "Move widget '".concat(w.name, "' to container '").concat(dest ? dest.name : '[forest]', "' at position ").concat(addAt),
                  redo: function redo() {
                    if (source) {
                      source.remove(removeAt);

                      if (source.delegate && (source.delegate instanceof CardsPane__default["default"] || source.delegate instanceof TabbedPane__default["default"])) {
                        w.delegate.visible = true;
                      }
                    } else {
                      self.forest.splice(removeAt, 1);
                    }

                    if (dest) {
                      dest.add(w, addAt);
                    } else {
                      self.forest.splice(addAt, 0, w);
                    }

                    self.layout.explorer.removed(w);
                    self.layout.explorer.added(w);
                    self.layout.explorer.goTo(w, true);

                    if (!(w.delegate instanceof ColumnNode__default["default"])) {
                      locationSize.applySizeLocationSnapshot(newLocationSize, w.delegate);
                    }
                  },
                  undo: function undo() {
                    if (dest) {
                      dest.remove(addAt);

                      if (dest.delegate && (dest.delegate instanceof CardsPane__default["default"] || dest.delegate instanceof TabbedPane__default["default"])) {
                        w.delegate.visible = true;
                      }
                    } else {
                      self.forest.splice(addAt, 1);
                    }

                    if (source) {
                      source.add(w, removeAt);
                    } else {
                      self.forest.splice(removeAt, 0, w);
                    }

                    self.layout.explorer.removed(w);
                    self.layout.explorer.added(w);
                    self.layout.explorer.goTo(w, true);

                    if (!(w.delegate instanceof ColumnNode__default["default"])) {
                      locationSize.applySizeLocationSnapshot(oldLocationSize, w.delegate);
                    }
                  }
                };
                self.edit(ur, parentUR);
                return true;
              } else {
                Logger__default["default"].info("Widget '".concat(w.name, "' is a child of '").concat(dest ? dest.name : '[forest]', "' already at the same position."));
              }
            } else {
              Logger__default["default"].info("Can't add widget '".concat(w.name, "' to widget '").concat(dest ? dest.name : '[forest]', "'. ").concat(dest ? dest.name : '[forest]', "' is full and can't accept more children."));
            }
          } else {
            Logger__default["default"].info("Can't add widget '".concat(w.name, "' to widget '").concat(dest ? dest.name : '[forest]', "'. Widget '").concat(w.name, "' is parent of '").concat(dest ? dest.name : '[forest]', "'."));
          }
        } else {
          Logger__default["default"].info("Can't add widget '".concat(w.name, "' to itself.'"));
        }
      } else {
        Logger__default["default"].info("Can't add '".concat(w.name, "' to '").concat(dest.name, "'. Widget '").concat(dest.name, "' should be a container and '").concat(w.name, "' should be a widget."));
      }

      return false;
    }

    this.move = move;

    this.layout.explorer.onDropInto = function (w, dest) {
      (self.layout.explorer.isSelected(w) ? self.layout.explorer.selected : [w]).forEach(function (m) {
        return move(m, dest, dest.count);
      });
    };

    this.layout.explorer.onDropBefore = function (w, before) {
      var dest = before.parent;
      var addAt = dest ? dest.indexOf(before) : self.forest.indexOf(before);
      (self.layout.explorer.isSelected(w) ? self.layout.explorer.selected : [w]).reverse().forEach(function (m) {
        return move(m, dest, addAt);
      });
    };

    this.layout.explorer.onDropAfter = function (w, after) {
      var dest = after.parent;
      var addAt = dest ? dest.indexOf(after) : self.forest.indexOf(after);
      (self.layout.explorer.isSelected(w) ? self.layout.explorer.selected : [w]).reverse().forEach(function (m) {
        return move(m, dest, addAt + 1);
      });
    };

    rewidth(this.layout.paletteExplorerSplit, self.layout.leftSizer, this.layout.view, 1);
    rewidth(this.layout.propertiesBox, self.layout.rightSizer, this.layout.view, -1);
    this.layout.propNameColumn.field = 'name';
    this.layout.propValueColumn.editor = new textNumberField();
    this.layout.propValueColumn.field = 'value';

    this.layout.propNameColumn.onRender = function (item, viewCell) {
      viewCell.title = item.name;
      viewCell.innerText = item.name;

      if (item.edited) {
        viewCell.classList.add('p-winnie-property-edited');
      }
    };

    this.layout.propValueColumn.onRender = render;
    enabled.push(function () {
      self.layout.miToSurface.enabled = self._lastSelected && !self._lastSelected.delegate.parent && self._lastSelected.delegate instanceof Container__default["default"];
    });

    this.layout.miToSurface.onAction = function () {
      checkEnabled();
      self.acceptVisualRoot(self._lastSelected);
      self.centerSurface();
      self.layout.explorer.changed(self._lastSelected);
    };

    this.layout.tOpen.onAction = function () {
      checkEnabled();
      self.openJSON();
    };

    this.layout.tTemplates.onAction = function () {
      checkEnabled();
      self.layout.templatesMenu.popupRelativeTo(self.layout.tTemplates.element, false);
    };

    this.layout.tSave.element.onclick = function () {
      checkEnabled();
      self.save();
    };

    this.layout.tExport.element.onclick = function () {
      checkEnabled();
      self["export"]();
    };

    [this.layout.tCut, this.layout.miCut].forEach(function (w) {
      enabled.push(function () {
        w.enabled = self.layout.explorer.selected.length > 0;
      });

      w.onAction = function () {
        checkEnabled();
        self.cut();
      };
    });
    [this.layout.tCopy, this.layout.miCopy].forEach(function (w) {
      enabled.push(function () {
        w.enabled = self.layout.explorer.selected.length > 0;
      });

      w.onAction = function () {
        checkEnabled();
        self.copy();
      };
    });
    [this.layout.tPaste, this.layout.miPaste].forEach(function (w) {
      Ui__default["default"].on(w.element, Ui__default["default"].Events.CLICK, function () {
        checkEnabled();
        self.alert(i18n['winnie.paste.via.shortcut']);
      });
    });
    [this.layout.tUndo, this.layout.miUndo].forEach(function (w) {
      enabled.push(function () {
        w.enabled = self.canUndo;
      });

      w.onAction = function () {
        checkEnabled();
        self.undo();
      };
    });
    [this.layout.tRedo, this.layout.miRedo].forEach(function (w) {
      enabled.push(function () {
        w.enabled = self.canRedo;
      });

      w.onAction = function () {
        checkEnabled();
        self.redo();
      };
    });
    [this.layout.tRemove, this.layout.miRemove].forEach(function (w) {
      enabled.push(function () {
        w.enabled = self.layout.explorer.selected.length > 0;
      });

      w.onAction = function () {
        checkEnabled();
        self.removeSelected();
      };
    });

    this.layout.tSettings.onAction = function () {
      checkEnabled();
      self.openSettings();
    };

    checkEnabled();
    this._palette = {};
    this._adopts = [];

    this.alert = function (msg) {
      alert(msg);
    };
  }

  _createClass__default["default"](Winnie, [{
    key: "save",
    value: function save() {
      if (this.widgets.size > 0) {
        var generatedCode = modelToCode(this);
        clipboard.write(generatedCode);
        this.layout.widgets.element.focus();
        Logger__default["default"].info(i18n['winnie.generated.es6.copied']);
        this.alert(i18n['winnie.generated.es6.copied']);
      } else {
        Logger__default["default"].info("Can't generate code for an empty [forest].");
      }
    }
  }, {
    key: "clear",
    value: function clear() {
      this.clearVisualRoot();
      this.editsCursor = 0;
      this.edits = [];
      this.widgets.clear();
      var removed = this.forest.splice(0, this.forest.length);
      this.layout.explorer.unselectAll();
      this.layout.explorer.removed(removed);
      this.layout.properties.data = [];
    }
  }, {
    key: "acceptJson",
    value: function acceptJson(parsed, initialParent) {
      var undoable = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
      var self = this;
      var indexed = {};
      Object.keys(self._palette).forEach(function (c) {
        var category = self._palette[c];
        category.items.forEach(function (item) {
          indexed[item.from] = item;
        });
      });
      var addLog = [];

      function adopt(plain, parent) {
        Object.keys(plain).forEach(function (plainName) {
          var item = plain[plainName];
          var widgetName = self.generateName(plainName);
          var source = indexed[item.from];

          if (source) {
            var created = new winnieWidget(widgets.produced(self, new source.widget()), widgetName, source.defaultInstance, function (newName) {
              widgets.rename(self, created, newName);
            });
            created.source = source;
            self.widgets.set(widgetName, created);

            if (parent) {
              addLog.push({
                subject: created,
                parent: parent,
                at: parent.count
              });
              parent.add(created, parent.count);
            } else {
              addLog.push({
                subject: created,
                parent: null,
                at: self.forest.length
              });
              self.forest.push(created);
            }

            for (var p in item.body) {
              if (p.includes('.')) {
                Bound__default["default"].setPathData(created.delegate, p, item.body[p]);
              } else {
                created.delegate[p] = item.body[p];
              }
            }

            created.sheet = properties(self, created);
            adopt(item.children, created);
          } else {
            Logger__default["default"].info("Can't read widget '".concat(widgetName, "' from an unknown module '").concat(item.from, "'."));
          }
        });
      }

      self.layout.explorer.unselectAll();
      adopt(parsed, initialParent);

      if (undoable) {
        var editBody = {
          name: "Paste ".concat(addLog.length === 1 ? "'".concat(addLog[0].subject.name, "' widget") : "".concat(addLog.length, " widgets"), " from clipboard"),
          redo: function redo() {},
          undo: function undo() {
            self.layout.explorer.unselectAll();
            addLog.slice(0, addLog.length).reverse().forEach(function (item) {
              self.widgets["delete"](item.subject.name);

              if (item.parent) {
                item.parent.remove(item.at);
              } else {
                self.forest.splice(item.at, 1);
              }
            });
            self.layout.explorer.removed(addLog.map(function (item) {
              return item.subject;
            }));
          }
        };
        this.edit(editBody);

        editBody.redo = function () {
          self.layout.explorer.unselectAll();
          addLog.forEach(function (item) {
            self.widgets.set(item.subject.name, item.subject);

            if (item.parent) {
              item.parent.add(item.subject, item.at);
            } else {
              self.forest.splice(item.at, 0, item.subject);
            }
          });
          self.layout.explorer.added(addLog.map(function (item) {
            return item.subject;
          }));
        };
      }

      self.layout.explorer.added(addLog.map(function (item) {
        return item.subject;
      }));
      return addLog;
    }
  }, {
    key: "acceptNatives",
    value: function acceptNatives(natives) {
      var self = this;
      var indexed = {};
      Object.keys(self._palette).forEach(function (c) {
        var category = self._palette[c];
        category.items.forEach(function (item) {
          indexed[item.widget] = item;
        });
      });

      function adopt(natives) {
        return Object.keys(natives).map(function (widgetName) {
          var _native = natives[widgetName];
          var source = indexed[_native.constructor];

          if (source) {
            var _created = new winnieWidget(widgets.produced(self, _native), widgetName, source.defaultInstance, function (newName) {
              widgets.rename(self, _created, newName);
            });

            _created.source = source;
            self.widgets.set(widgetName, _created);

            if (_native instanceof ColumnNode__default["default"]) {
              if (!_native.column || !_native.column.grid) {
                self.forest.push(_created);
              }
            } else if (!_native.parent) {
              self.forest.push(_created);
            }

            _created.sheet = properties(self, _created);
            return _created;
          } else {
            Logger__default["default"].info("Can't read widget '".concat(widgetName, "' because its constructor is not registered."));
            return null;
          }
        }).filter(function (item) {
          return !!item;
        });
      }

      var created = adopt(natives);
      this.layout.explorer.added(this.forest);
      return created;
    }
  }, {
    key: "generateName",
    value: function generateName(base) {
      var widgetName = base;
      var nameAttempt = 1;

      while (this.widgets.has(widgetName)) {
        widgetName = "".concat(base).concat(nameAttempt++);
      }

      return widgetName;
    }
  }, {
    key: "walk",
    value: function walk(widget, onEach) {
      var self = this;
      onEach(widget);
      widget.children.forEach(function (child) {
        return self.walk(child, onEach);
      });
    }
  }, {
    key: "addWidget",
    value: function addWidget(item) {
      var left = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;
      var top = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
      var self = this;
      var wasSelected = this._lastSelected;
      var widgetName = this.generateName(decapitalize(classes(item.defaultInstance)));
      var created = new winnieWidget(widgets.produced(self, widgets.produce(item.widget, widgetName, item.name, "".concat(self.settings.grid.x, "px"), "".concat(self.settings.grid.y, "px"))), widgetName, item.defaultInstance, function (newName) {
        widgets.rename(self, created, newName);
      });
      created.source = item;
      var addEdit = {
        name: "Add '".concat(item.name, "' widget from '").concat(item.from, "'"),
        redo: function redo() {
          self.widgets.set(widgetName, created);
          self.forest.push(created);
          self.layout.explorer.added(created);
          self.layout.explorer.goTo(created, true);
        },
        undo: function undo() {
          self.widgets["delete"](widgetName);
          var createdIndex = self.forest.indexOf(created);
          var removed = self.forest.splice(createdIndex, 1);
          self.layout.explorer.unselect(created);
          self.layout.explorer.removed(created);
          self.revokeVisualRoot(created);
        }
      };
      this.edit(addEdit);

      if (created.delegate instanceof Container__default["default"] && !created.parent && !this.visualRootPresent()) {
        this.acceptVisualRoot(created);

        if (created.delegate instanceof Container__default["default"]) {
          created.delegate.width = created.delegate.height = 300;
        }

        this.centerSurface();
      }

      var initSize = function initSize() {
        if (created.delegate instanceof Container__default["default"] && !(created.delegate instanceof Menu__default["default"]) || created.delegate instanceof DataGrid__default["default"]) {
          if (left != null) {
            created.delegate.left = Math.max(0, left - 64);
          }

          if (top != null) {
            created.delegate.top = Math.max(0, top - 64);
          }

          created.delegate.width = 128;
          created.delegate.height = 128;
        } else if (!(created.delegate instanceof MenuElement__default["default"])) {
          if (left != null) {
            created.delegate.left = Math.max(0, left - 48);
          }

          if (top != null) {
            created.delegate.top = Math.max(0, top - 12);
          } //created.delegate.height = 32;
          //created.delegate.width = 96;

        }
      };

      if (created.delegate instanceof DataGrid__default["default"]) {
        created.delegate.showOddRowsInOtherColor = false;
      }

      if (wasSelected) {
        if (wasSelected.delegate instanceof Container__default["default"] || wasSelected.delegate instanceof DataGrid__default["default"]) {
          if (wasSelected.delegate instanceof Anchors__default["default"]) {
            initSize();
          }

          this.move(created, wasSelected, wasSelected.count, addEdit);
        } else if (wasSelected.parent && (wasSelected.parent.delegate instanceof Container__default["default"] || wasSelected.parent.delegate instanceof DataGrid__default["default"])) {
          if (wasSelected.parent.delegate instanceof Anchors__default["default"]) {
            initSize();
          }

          this.move(created, wasSelected.parent, wasSelected.parent.count, addEdit);
        }
      }

      this.checkEnabled();
      return created;
    }
  }, {
    key: "removeSelected",
    value: function removeSelected() {
      var self = this;
      var toRemove = self.layout.explorer.selected;

      if (toRemove.length > 0) {
        var added = function added(item, wasVisualRoot) {
          self.layout.explorer.unselectAll();
          self.layout.explorer.select(item);
          self.layout.explorer.added(item);

          if (wasVisualRoot) {
            self.acceptVisualRoot(item);
          }
        };

        var removed = function removed(item) {
          self.layout.explorer.unselect(item);
          self.layout.explorer.removed(item);
          return self.revokeVisualRoot(item);
        };

        var firstSelectedParent = null;
        var firstSelectedAt = -1;
        var viewRows = self.layout.explorer.viewRows;

        for (var i = 0; i < viewRows.length; i++) {
          var widget = viewRows[i];

          if (self.layout.explorer.isSelected(widget)) {
            firstSelectedParent = widget.parent;
            firstSelectedAt = firstSelectedParent ? firstSelectedParent.indexOf(widget) : self.forest.indexOf(widget);
            break;
          }
        }

        var actions = toRemove.map(function (item) {
          var itemParent = item.parent;
          var removedAt = itemParent ? itemParent.indexOf(item) : self.forest.indexOf(item);
          var wasVisualRoot = null;
          return {
            removedAt: removedAt,
            redo: function redo() {
              if (itemParent) {
                itemParent.remove(removedAt);
              } else {
                self.forest.splice(removedAt, 1);
              }

              self.walk(item, function (_item) {
                self.widgets["delete"](_item.name);
              });
              wasVisualRoot = removed(item);
            },
            undo: function undo() {
              self.walk(item, function (_item) {
                self.widgets.set(_item.name, _item);
              });

              if (itemParent) {
                itemParent.add(item, removedAt);
                added(item);
              } else {
                self.forest.splice(removedAt, 0, item);
                added(item, wasVisualRoot);
              }
            }
          };
        }).sort(function (action1, action2) {
          return action2.removedAt - action1.removedAt;
        });
        this.edit({
          name: "Delete ".concat(toRemove.length > 1 ? "selected (".concat(toRemove.length, ") widgets") : "'".concat(toRemove[0].name, "' widget")),
          redo: function redo() {
            actions.forEach(function (item) {
              item.redo();
            });
          },
          undo: function undo() {
            actions.slice(0, actions.length).reverse().forEach(function (item) {
              item.undo();
            });
          }
        });
        this.checkEnabled();

        while (firstSelectedAt > -1) {
          var toSelect = null;

          if (firstSelectedParent) {
            var firstParentChildren = firstSelectedParent.children;

            if (firstParentChildren.length > 0) {
              toSelect = firstSelectedAt >= 0 && firstSelectedAt < firstParentChildren.length ? firstParentChildren[firstSelectedAt] : null;
            } else {
              toSelect = firstSelectedParent;
            }
          } else {
            toSelect = firstSelectedAt < self.forest.length ? self.forest[firstSelectedAt] : null;
          }

          if (toSelect) {
            self.layout.explorer.goTo(toSelect, true);
            break;
          } else {
            firstSelectedAt--;
          }
        }
      }
    }
  }, {
    key: "copy",
    value: function copy() {
      if (this.layout.explorer.selected.length > 0) {
        var selected = new Set(this.layout.explorer.selected);
        var toCopy = Array.from(selected).filter(function (s) {
          var w = s;

          while (w.parent) {
            w = w.parent;

            if (selected.has(w)) {
              return false;
            }
          }

          return true;
        });
        var generatedJson = modelToJson(toCopy);
        clipboard.write(generatedJson);
        this.layout.widgets.element.focus();
        Logger__default["default"].info("Copied ".concat(toCopy.length === 1 ? "'".concat(toCopy[0].name, "' widget") : "".concat(toCopy.length, " widgets"), " to clipboard."));
      }
    }
  }, {
    key: "paste",
    value: function paste(source) {
      var self = this;

      if (source) {
        var initialParent = this.lastSelected ? this.lastSelected.delegate instanceof Container__default["default"] ? this.lastSelected : this.lastSelected.parent : null;
        var created = this.acceptJson(JSON.parse(source), initialParent, true);
        created.forEach(function (item) {
          if (item.subject.delegate.parent instanceof Anchors__default["default"]) {
            item.subject.delegate.left += self.settings.grid.x;
            item.subject.delegate.top += self.settings.grid.y;
          }

          self.layout.explorer.select(item.subject);
        });
        return created;
      } else {
        return [];
      }
    }
  }, {
    key: "cut",
    value: function cut() {
      if (this.layout.explorer.selected.length > 0) {
        this.copy();
        this.removeSelected();
      }
    }
  }, {
    key: "openJSON",
    value: function openJSON() {
      var self = this;
      var input = document.createElement('input');
      input.type = 'file';
      input.multiple = false;
      input.accept = '.json';
      input.style.width = input.style.height = '2em';
      input.style.top = input.style.left = '-1000px';
      document.body.appendChild(input);
      Ui__default["default"].on(input, Ui__default["default"].Events.CHANGE, function () {
        if (input.value) {
          var reader = new FileReader();

          reader.onload = function () {
            self.openParsedJson(JSON.parse(reader.result));
          };

          reader.readAsText(input.files[0]);
        }

        document.body.removeChild(input);
      });
      Ui__default["default"].delayed(60000, function () {
        if (input.parentElement) {
          document.body.removeChild(input);
        }
      });
      input.click();
    }
  }, {
    key: "openParsedJson",
    value: function openParsedJson(parsed) {
      var self = this;
      var firstContainer = null;
      self.clear();
      var addLog = self.acceptJson(parsed);
      addLog.forEach(function (item) {
        if (!firstContainer && item.subject.delegate instanceof Container__default["default"] && !item.subject.delegate.parent) {
          firstContainer = item.subject;
        }
      });

      if (firstContainer) {
        this.acceptVisualRoot(firstContainer);
        this.layout.explorer.goTo(firstContainer, true);
        this.centerSurface();
      }
    }
  }, {
    key: "openNatives",
    value: function openNatives(Natives) {
      var firstContainer = null;
      this.clear();
      this.acceptNatives(new Natives()).forEach(function (item) {
        if (!firstContainer && item.delegate instanceof Container__default["default"] && !item.delegate.parent) {
          firstContainer = item;
        }
      });

      if (firstContainer) {
        this.acceptVisualRoot(firstContainer);
        this.layout.explorer.goTo(firstContainer, true);
        this.centerSurface();
      }

      this.layout.widgets.element.focus();
    }
  }, {
    key: "export",
    value: function _export() {
      if (this.widgets.size > 0) {
        var generatedJson = modelToJson(this.forest);
        clipboard.write(generatedJson);
        this.layout.widgets.element.focus();
        Logger__default["default"].info(i18n['winnie.generated.json.copied']);
        this.alert(i18n['winnie.generated.json.copied']);
      } else {
        Logger__default["default"].info("Can't generate JSON for an empty [forest].");
      }
    }
  }, {
    key: "openSettings",
    value: function openSettings() {}
  }, {
    key: "canRedo",
    get: function get() {
      return this.editsCursor >= 0 && this.editsCursor < this.edits.length;
    }
  }, {
    key: "redo",
    value: function redo() {
      if (this.canRedo) {
        var item = this.edits[this.editsCursor];

        while (item) {
          item.redo();
          Logger__default["default"].info("Redone edit: ".concat(item.name, "."));
          item = item.subEdit;
        }

        this.editsCursor++;
        this.checkEnabled();

        if (this.editsCursor > this.edits.length) {
          debugger;
        }
      }
    }
  }, {
    key: "canUndo",
    get: function get() {
      return this.editsCursor > 0 && this.editsCursor <= this.edits.length;
    }
  }, {
    key: "undo",
    value: function undo() {
      if (this.canUndo) {
        --this.editsCursor;
        var toUndo = [];
        var item = this.edits[this.editsCursor];

        while (item) {
          toUndo.unshift(item);
          item = item.subEdit;
        }

        toUndo.forEach(function (ur) {
          ur.undo();
          Logger__default["default"].info("Undone edit: ".concat(ur.name, "."));
        });
        this.checkEnabled();

        if (this.editsCursor > this.edits.length) {
          debugger;
        }
      }
    }
  }, {
    key: "edit",
    value: function edit(body, parentUR) {
      body.redo();

      if (parentUR) {
        parentUR.subEdit = body;
      } else {
        var dropped = this.edits.splice(this.editsCursor, this.edits.length - this.editsCursor, body);

        if (dropped.length > 0) {
          Logger__default["default"].info("Dropped ".concat(dropped.length, " tail edits."));
        }

        this.editsCursor++;

        while (this.edits.length > this.settings.undoDepth) {
          var shifted = this.edits.shift();
          this.editsCursor--;
          Logger__default["default"].info("Dropped head edit: '".concat(shifted.name, "'."));
        }

        if (this.editsCursor > this.edits.length) {
          debugger;
        }
      }

      Logger__default["default"].info("Recorded edit: ".concat(body.name, "."));
    }
  }, {
    key: "palette",
    get: function get() {
      var self = this;

      function category(name) {
        var box = new Box__default["default"](Ui__default["default"].Orientation.VERTICAL);
        var header = new Label__default["default"](name);
        var headerIcon = document.createElement('div');
        headerIcon.className = 'winnie-icon-down-open';
        header.icon = headerIcon;
        header.element.classList.add('p-widget-category-header');
        var content = new FlowPane__default["default"](5, 5);
        content.element.classList.add('p-widget-category-content');
        box.add(header);
        box.add(content);
        self.layout.palette.add(box);

        header.onMouseClick = function (evt) {
          if (headerIcon.classList.contains('winnie-icon-down-open')) {
            headerIcon.classList.remove('winnie-icon-down-open');
            headerIcon.classList.add('winnie-icon-right-open');
            content.height = 0;
          } else {
            headerIcon.classList.remove('winnie-icon-right-open');
            headerIcon.classList.add('winnie-icon-down-open');
            content.height = null;
          }
        };

        var menuItem = new MenuItem__default["default"](name);
        var menuIcon = document.createElement('div');
        menuIcon.className = 'winnie-icon-space';
        menuItem.icon = menuIcon;
        menuItem.subMenu = new Menu__default["default"]();
        self.layout.miAdd.subMenu.add(menuItem);
        return {
          name: name,
          palette: {
            box: box,
            header: header,
            content: content
          },
          menu: menuItem.subMenu,
          items: []
        };
      }

      function categoryOf(item) {
        var categoryName;

        if (item.category) {
          categoryName = item.category;
        } else {
          categoryName = i18n['winnie.category.default'];
          Logger__default["default"].info("Palette item from '".concat(item.from, "' hs no category, so it goers to default one."));
        }

        return categoryName in self._palette ? self._palette[categoryName] : self._palette[categoryName] = category(categoryName);
      }

      function paletteItemOf(item) {
        var itemLabel = new Label__default["default"](item.name);
        itemLabel.horizontalTextPosition = Ui__default["default"].HorizontalPosition.CENTER;
        itemLabel.verticalTextPosition = Ui__default["default"].VerticalPosition.BOTTOM;
        var div = document.createElement('div');
        div.className = item.iconStyle ? item.iconStyle : 'winnie-icon-wrench';
        itemLabel.icon = div;
        itemLabel.opaque = true;
        itemLabel.element.classList.add('p-widget-palette-item');

        if ('description' in item) {
          itemLabel.toolTipText = item.description;
        }

        itemLabel.element.draggable = true;
        Ui__default["default"].on(itemLabel.element, Ui__default["default"].Events.DRAGSTART, function (event) {
          event.stopPropagation();
          self.paletteDrag = {
            item: item
          };
          event.dataTransfer.effectAllowed = 'move';
          event.dataTransfer.setData('text/plain', 'p-winnie-palette-item');
        });
        Ui__default["default"].on(itemLabel.element, Ui__default["default"].Events.DRAGEND, function (event) {
          event.stopPropagation();

          if (self.paletteDrag) {
            self.paletteDrag = null;
          }
        });
        return itemLabel;
      }

      function menuItemOf(item) {
        var itemMi = new MenuItem__default["default"](item.name);
        var div = document.createElement('div');
        div.className = item.iconStyle ? item.iconStyle : 'winnie-icon-wrench';
        itemMi.icon = div;
        return itemMi;
      }

      return {
        add: function add(items) {
          if (!Array.isArray(items)) items = [items];
          items.filter(function (item) {
            if (!item.name) {
              Logger__default["default"].info("Can't add anonymous palette item from '".concat(item.from, "'. Skipping..."));
              return false;
            } else if (!item.from) {
              Logger__default["default"].info("Can't add palette item '".concat(item.name, "' without 'from' field. Skipping..."));
              return false;
            } else if (!item.widget) {
              Logger__default["default"].info("Can't add palette item '".concat(item.name, "' without 'widget' field. Skipping..."));
              return false;
            } else {
              return true;
            }
          }).map(function (item) {
            return {
              widget: item.widget,
              from: item.from,
              name: item.name,
              description: item.description,
              iconStyle: item.iconStyle,
              category: item.category,
              hidden: new Set(Array.isArray(item.hidden) ? item.hidden : []),
              selects: item.selects ? item.selects : {}
            };
          }).forEach(function (item) {
            try {
              item.defaultInstance = new item.widget();
              var itemPaletteItem = paletteItemOf(item);
              var itemMenuItem = menuItemOf(item);

              itemMenuItem.onAction = function () {
                self.addWidget(item);
              };

              var _category = categoryOf(item);

              _category.palette.content.add(itemPaletteItem);

              _category.menu.add(itemMenuItem);

              _category.items.push(item);
            } catch (ex) {
              Logger__default["default"].info("Can't add palette item '".concat(item.name, "' from '").concat(item.from, "' due to exception:"));
              Logger__default["default"].severe(ex);
              Logger__default["default"].info("Palette item '".concat(item.name, "' from '").concat(item.from, "' skipped."));
            }
          });
        }
      };
    }
  }, {
    key: "adopts",
    get: function get() {
      var self = this;
      return {
        add: function add(items) {
          if (!Array.isArray(items)) items = [items];
          items.filter(function (item) {
            if (!item.name) {
              Logger__default["default"].info("Can't add anonymous module. Skipping...");
              return false;
            } else if (!item.widgets) {
              Logger__default["default"].info("Can't add module '".concat(item.name, "' without 'widgets' field. Skipping..."));
              return false;
            } else {
              return true;
            }
          }).sort(function (item1, item2) {
            if (item1.name > item2.name) {
              return 1;
            } else if (item1.name < item2.name) {
              return -1;
            } else {
              return 0;
            }
          }).forEach(function (item) {
            var mi = new MenuItem__default["default"](item.name);

            mi.onAction = function () {
              self.openNatives(item.widgets);
            };

            self.layout.templatesMenu.add(mi);
          });
        }
      };
    }
  }, {
    key: "lastSelected",
    get: function get() {
      return this._lastSelected;
    },
    set: function set(item) {
      var self = this;
      this.widgets.forEach(function (item, name) {
        if (item.delegate instanceof Widget__default["default"]) {
          item.delegate.element.classList.remove('p-winnie-widget-selected');
        }
      });
      this.layout.explorer.selected.filter(function (item) {
        return item.delegate instanceof Widget__default["default"];
      }).forEach(function (selectedItem) {
        selectedItem.delegate.element.classList.add('p-winnie-widget-selected');
      });
      self.undecorate();

      if (this._lastSelected !== item) {
        if (this._lastSelected) {
          this.layout.properties.data = [];
          self.checkEnabled();
        }

        this._lastSelected = item;

        if (this._lastSelected) {
          var sheet;

          if (item.sheet) {
            sheet = item.sheet;
          } else {
            item.sheet = sheet = properties(self, item);
          }

          this.layout.properties.data = sheet.filter(function (p) {
            return !p.name.startsWith('tab.') || item.delegate.parent instanceof TabbedPane__default["default"];
          });
          self.checkEnabled();
        }
      }

      self.decorate();
    }
  }, {
    key: "decorate",
    value: function decorate() {
      this.undecorate();

      if (this._lastSelected && this.layout.explorer.selected.length === 1) {
        var self = this;
        var subject = self._lastSelected;

        self.decors = function () {
          if (subject.delegate.attached) {
            var decors = locationSize.resizeDecor(self.layout.widgets, subject.delegate, function (prevState, diff) {
              var subjectElement = subject.delegate.element;

              if (diff.x !== 0 || diff.y !== 0) {
                var newState = {
                  anchors: {
                    left: subjectElement.style.left,
                    width: subjectElement.style.width,
                    right: subjectElement.style.right,
                    top: subjectElement.style.top,
                    height: subjectElement.style.height,
                    bottom: subjectElement.style.bottom
                  }
                };
                self.edit({
                  name: "Widget '".concat(subject.name, "' resize"),
                  redo: function redo() {
                    subjectElement.style.left = newState.anchors.left;
                    subjectElement.style.width = newState.anchors.width;
                    subjectElement.style.right = newState.anchors.right;
                    subjectElement.style.top = newState.anchors.top;
                    subjectElement.style.height = newState.anchors.height;
                    subjectElement.style.bottom = newState.anchors.bottom;
                    self.stickDecors();
                    self.layout.explorer.goTo(subject, true);
                  },
                  undo: function undo() {
                    subjectElement.style.left = prevState.anchors.left;
                    subjectElement.style.width = prevState.anchors.width;
                    subjectElement.style.right = prevState.anchors.right;
                    subjectElement.style.top = prevState.anchors.top;
                    subjectElement.style.height = prevState.anchors.height;
                    subjectElement.style.bottom = prevState.anchors.bottom;
                    self.stickDecors();
                    self.layout.explorer.goTo(subject, true);
                  }
                });
              }
            });

            if (subject.delegate.parent instanceof Anchors__default["default"]) {
              return [decors.lt, decors.lm, decors.lb, decors.mt, decors.mb, decors.rt, decors.rm, decors.rb];
            } else if (subject.delegate.parent instanceof FlowPane__default["default"] || subject.delegate.parent instanceof ScrollPane__default["default"]) {
              return [decors.mb, decors.rm, decors.rb];
            } else if (subject.delegate.parent instanceof Box__default["default"] && subject.delegate.parent.orientation === Ui__default["default"].Orientation.VERTICAL) {
              return [decors.mb];
            } else if (subject.delegate.parent instanceof Box__default["default"] && subject.delegate.parent.orientation === Ui__default["default"].Orientation.HORIZONTAL || subject.delegate.parent instanceof Toolbar__default["default"]) {
              return [decors.rm];
            } else if (subject.delegate.parent instanceof HolyGrailPane__default["default"] && subject.delegate.parent.header === subject.delegate) {
              return [decors.mb];
            } else if (subject.delegate.parent instanceof HolyGrailPane__default["default"] && subject.delegate.parent.leftSide === subject.delegate) {
              return [decors.rm];
            } else if (subject.delegate.parent instanceof HolyGrailPane__default["default"] && subject.delegate.parent.rightSide === subject.delegate) {
              return [decors.lm];
            } else if (subject.delegate.parent instanceof HolyGrailPane__default["default"] && subject.delegate.parent.footer === subject.delegate) {
              return [decors.mt];
            } else {
              return [];
            }
          } else {
            return [];
          }
        }().map(function (d, i, source) {
          self.layout.widgets.element.appendChild(d);
          d.neightbours = source;
          d.stick();
          return d;
        });
      }
    }
  }, {
    key: "undecorate",
    value: function undecorate() {
      this.decors.forEach(function (d) {
        if (d.parentElement) {
          d.parentElement.removeChild(d);
        }
      });
      this.decors = [];
    }
  }, {
    key: "visualRootElement",
    value: function visualRootElement() {
      var child = this.layout.widgets.element.firstElementChild;

      while (child) {
        if (child.tagName.toLowerCase() !== 'style' && !child.className.includes('p-winnie-decoration') && !child.className.includes('p-winnie-hint')) {
          return child;
        }

        child = child.nextElementSibling;
      }

      return null;
    }
  }, {
    key: "visualRootPresent",
    value: function visualRootPresent() {
      return !!this.visualRootElement();
    }
  }, {
    key: "acceptVisualRoot",
    value: function acceptVisualRoot(w) {
      if (!w.delegate.parent) {
        // Only roots of widgets heirarchy can be accepted as visual root of designer
        var oldVRE = this.visualRootElement();

        if (oldVRE !== w.delegate.element) {
          this.clearVisualRoot();
          this.layout.widgets.element.appendChild(w.delegate.element);
        }
      }
    }
  }, {
    key: "clearVisualRoot",
    value: function clearVisualRoot() {
      var vre = this.visualRootElement();

      if (vre) {
        vre.parentElement.removeChild(vre);
      }
    }
  }, {
    key: "revokeVisualRoot",
    value: function revokeVisualRoot(w) {
      if (w.delegate instanceof Container__default["default"] && w.delegate.element.parentElement === this.layout.widgets.element) {
        this.layout.widgets.element.removeChild(w.delegate.element);
        return true;
      } else {
        return false;
      }
    }
  }, {
    key: "centerSurface",
    value: function centerSurface() {
      this.layout.view.element.scrollLeft = (this.layout.widgets.element.offsetWidth - this.layout.view.element.clientWidth) / 2;
      this.layout.view.element.scrollTop = (this.layout.widgets.element.offsetHeight - this.layout.view.element.clientHeight) / 2;
      this.layout.widgets.element.focus();
    }
  }, {
    key: "stickDecors",
    value: function stickDecors() {
      this.decors.forEach(function (d) {
        d.stick();
      });
    }
  }]);

  return Winnie;
}();

module.exports = Winnie;
