'use strict';

var _toConsumableArray = require('@babel/runtime/helpers/toConsumableArray');
var Ui = require('kenga/utils');
var Bound = require('kenga/bound');
var Widget = require('kenga/widget');
var BoxField = require('kenga/box-field');
var ColumnNode = require('kenga-grid/columns/column-node');
var info = require('../properties/info.js');
var winnieProperty = require('../properties/winnie-property.js');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var _toConsumableArray__default = /*#__PURE__*/_interopDefaultLegacy(_toConsumableArray);
var Ui__default = /*#__PURE__*/_interopDefaultLegacy(Ui);
var Bound__default = /*#__PURE__*/_interopDefaultLegacy(Bound);
var Widget__default = /*#__PURE__*/_interopDefaultLegacy(Widget);
var BoxField__default = /*#__PURE__*/_interopDefaultLegacy(BoxField);
var ColumnNode__default = /*#__PURE__*/_interopDefaultLegacy(ColumnNode);

function createProps(model, item) {
  var propNames = Object.getOwnPropertyNames(item.delegate);

  if (item.delegate instanceof Widget__default["default"]) {
    if (item.delegate instanceof Widget__default["default"] && !propNames.includes('classes')) {
      propNames.push('classes');
    }

    propNames.push.apply(propNames, _toConsumableArray__default["default"](info.pathProps));
  }

  return propNames.filter(function (key) {
    return typeof item.delegate[key] !== 'function' && (!(item.delegate instanceof Widget__default["default"]) || !info.widgetHidden.has(key)) && (!(item.delegate instanceof ColumnNode__default["default"]) || !info.dataGridColumnsHidden.has(key)) && (!(item.delegate instanceof BoxField__default["default"]) || !info.fieldsHidden.has(key)) && !item.source.hidden.has(key) && !key.startsWith('on');
  }).map(function (key) {
    var prop = new winnieProperty(item, key, function (newValue) {
      var oldValue = key === 'visible' ? prop.visible : key.includes('.') ? Bound__default["default"].getPathData(item.delegate, key) : item.delegate[key];
      var editBody = {
        name: "Property '".concat(key, "' of widget '").concat(item.name, "' change"),
        redo: function redo() {
          if (key === 'visible') {
            prop.visible = newValue;
          } else if (key === 'classes' && item.delegate instanceof Widget__default["default"]) {
            var element = item.delegate.element;

            if (oldValue) {
              oldValue.split(' ').forEach(function (className) {
                return element.classList.remove(className);
              });
            }

            item.delegate[key] = newValue;

            if (newValue) {
              newValue.split(' ').forEach(function (className) {
                return element.classList.add(className);
              });
            }
          } else {
            if (key.includes('.')) {
              Bound__default["default"].setPathData(item.delegate, key, newValue);
            } else {
              item.delegate[key] = newValue;
            }
          }

          if (!prop.silent && !model.layout.properties.activeEditor) {
            model.layout.properties.changed(prop);
            model.layout.properties.goTo(prop, true);
          }

          prop.silent = false;
          Ui__default["default"].delayed(10, function () {
            model.stickDecors();
          });
        },
        undo: function undo() {
          if (key === 'visible') {
            prop.visible = oldValue;
          } else if (key === 'classes' && item.delegate instanceof Widget__default["default"]) {
            var element = item.delegate.element;

            if (newValue) {
              newValue.split(' ').forEach(function (className) {
                return element.classList.remove(className);
              });
            }

            item.delegate[key] = oldValue;

            if (oldValue) {
              oldValue.split(' ').forEach(function (className) {
                return element.classList.add(className);
              });
            }
          } else {
            if (key.includes('.')) {
              Bound__default["default"].setPathData(item.delegate, key, oldValue);
            } else {
              item.delegate[key] = oldValue;
            }
          }

          if (!model.layout.properties.activeEditor) {
            model.layout.properties.changed(prop);
            model.layout.properties.goTo(prop, true);
          }

          Ui__default["default"].delayed(10, function () {
            model.stickDecors();
          });
        }
      };
      model.edit(editBody);
      model.checkEnabled();
    }, key.includes('.') ? Bound__default["default"].getPathData(item.defaultInstance, key) : item.defaultInstance[key]);

    if (item.delegate instanceof Widget__default["default"] && key === 'classes' && item.delegate[key]) {
      item.delegate.element.className += " ".concat(item.delegate[key]);
    }

    return prop;
  });
}

module.exports = createProps;
