'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var Ui = require('kenga/utils');
var Widget = require('kenga/widget');
var Container = require('kenga/container');
var ImageParagraph = require('kenga-labels/image-paragraph');
var CheckBox = require('kenga-buttons/check-box');
var RadioButton = require('kenga-buttons/radio-button');
var HorizontalBox = require('kenga-containers/horizontal-box-pane');
var VerticalBox = require('kenga-containers/vertical-box-pane');
var FlowPane = require('kenga-containers/flow-pane');
var GridPane = require('kenga-containers/grid-pane');
var Anchors = require('kenga-containers/anchors-pane');
var TabbedPane = require('kenga-containers/tabbed-pane');
var HolyGrailPane = require('kenga-containers/holy-grail-pane');
var CardsPane = require('kenga-containers/card-pane');
var MenuItem = require('kenga-menu/menu-item');
var DataGrid = require('kenga-grid/grid');
var ColumnNode = require('kenga-grid/columns/column-node');
var i18n = require('../i18n.js');
var locationSize = require('../location-size.js');
var rectSelection = require('../rect-selection.js');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var Ui__default = /*#__PURE__*/_interopDefaultLegacy(Ui);
var Widget__default = /*#__PURE__*/_interopDefaultLegacy(Widget);
var Container__default = /*#__PURE__*/_interopDefaultLegacy(Container);
var ImageParagraph__default = /*#__PURE__*/_interopDefaultLegacy(ImageParagraph);
var CheckBox__default = /*#__PURE__*/_interopDefaultLegacy(CheckBox);
var RadioButton__default = /*#__PURE__*/_interopDefaultLegacy(RadioButton);
var HorizontalBox__default = /*#__PURE__*/_interopDefaultLegacy(HorizontalBox);
var VerticalBox__default = /*#__PURE__*/_interopDefaultLegacy(VerticalBox);
var FlowPane__default = /*#__PURE__*/_interopDefaultLegacy(FlowPane);
var GridPane__default = /*#__PURE__*/_interopDefaultLegacy(GridPane);
var Anchors__default = /*#__PURE__*/_interopDefaultLegacy(Anchors);
var TabbedPane__default = /*#__PURE__*/_interopDefaultLegacy(TabbedPane);
var HolyGrailPane__default = /*#__PURE__*/_interopDefaultLegacy(HolyGrailPane);
var CardsPane__default = /*#__PURE__*/_interopDefaultLegacy(CardsPane);
var MenuItem__default = /*#__PURE__*/_interopDefaultLegacy(MenuItem);
var DataGrid__default = /*#__PURE__*/_interopDefaultLegacy(DataGrid);
var ColumnNode__default = /*#__PURE__*/_interopDefaultLegacy(ColumnNode);

function produce(constr, widgetName, paletteItemName, hgap, vgap) {
  var instance;

  if (constr === GridPane__default["default"]) {
    instance = new constr(hgap, vgap);
  } else if (constr === HorizontalBox__default["default"]) {
    instance = new constr(hgap);
  } else if (constr === VerticalBox__default["default"]) {
    instance = new constr(vgap);
  } else if (constr === FlowPane__default["default"] || constr === HolyGrailPane__default["default"]) {
    instance = new constr(hgap, vgap);
  } else if (constr === CardsPane__default["default"]) {
    instance = new constr();
    instance.hgap = hgap;
    instance.vgap = vgap;
  } else {
    instance = new constr();
  }

  if (instance instanceof ImageParagraph__default["default"] || instance instanceof MenuItem__default["default"] || instance instanceof CheckBox__default["default"] || instance instanceof RadioButton__default["default"]) {
    instance.text = widgetName.length > 1 ? widgetName.substring(0, 1).toUpperCase() + widgetName.substring(1) : widgetName;
  }

  if (instance instanceof ColumnNode__default["default"] && !instance.title) {
    instance.title = widgetName.length > 1 ? widgetName.substring(0, 1).toUpperCase() + widgetName.substring(1) : widgetName;
  }

  return instance;
}

function produced(model, instance) {
  function inRect(element, event) {
    var rect = element.getBoundingClientRect();
    return event.clientX >= rect.left && event.clientY >= rect.top && event.clientX < rect.right && event.clientY < rect.bottom;
  }

  if (instance instanceof Widget__default["default"]) {
    instance.element.classList.add('p-winnie-widget');
    Ui__default["default"].on(instance.element, Ui__default["default"].Events.MOUSEOVER, function () {
      instance.element.classList.add('p-winnie-widget-hover');
    });
    Ui__default["default"].on(instance.element, Ui__default["default"].Events.MOUSEOUT, function () {
      instance.element.classList.remove('p-winnie-widget-hover');
    });
    Ui__default["default"].on(instance.element, Ui__default["default"].Events.MOUSEDOWN, function (event) {
      var subject = instance['winnie.wrapper'];
      model.layout.explorer.goTo(subject);

      if (event.ctrlKey) {
        if (model.layout.explorer.isSelected(subject)) {
          model.layout.explorer.unselect(subject);
        } else {
          model.layout.explorer.select(subject);
        }
      } else {
        model.layout.explorer.unselectAll();
        model.layout.explorer.select(subject);
      }

      Ui__default["default"].later(function () {
        model.layout.widgets.element.focus();
      });
      event.stopPropagation();
    });
    locationSize.mouseDrag(instance.element, function (event) {
      if (instance.element === model.visualRootElement()) {
        return rectSelection.startRectSelection(model.layout.view.element, event);
      } else {
        return locationSize.startItemsMove(model, instance);
      }
    }, function (start, diff, event) {
      if (instance.element === model.visualRootElement()) {
        rectSelection.proceedRectSelection(start, diff);
      } else {
        locationSize.proceedItemsMove(model, start, diff, event.ctrlKey ? model.settings.grid.snap : false);
      }
    }, function (start, diff, event) {
      if (instance.element === model.visualRootElement()) {
        rectSelection.endRectSelection(model, start, diff, event);
        var subject = instance['winnie.wrapper'];
        model.layout.explorer.goTo(subject);
      } else {
        if (diff.x !== 0 || diff.y !== 0) {
          locationSize.endItemsMove(model, start);
        }
      }
    });

    if (instance instanceof Container__default["default"] || instance instanceof DataGrid__default["default"]) {
      Ui__default["default"].on(instance.element, Ui__default["default"].Events.DRAGENTER, function (event) {
        if (model.paletteDrag && (instance instanceof Container__default["default"] && event.target === instance.element || instance instanceof HolyGrailPane__default["default"] && event.target.parentElement === instance.element || instance instanceof TabbedPane__default["default"] && event.target.parentElement === instance.element || instance instanceof DataGrid__default["default"] && inRect(instance.element, event))) {
          event.preventDefault();
          event.stopPropagation();
          instance.element.classList.add('p-winnie-container-dnd-target');
        }
      });
      Ui__default["default"].on(instance.element, Ui__default["default"].Events.DRAGLEAVE, function (event) {
        if (model.paletteDrag && (instance instanceof Container__default["default"] && event.target === instance.element || instance instanceof HolyGrailPane__default["default"] && event.target.parentElement === instance.element || instance instanceof TabbedPane__default["default"] && event.target.parentElement === instance.element || instance instanceof DataGrid__default["default"] && !inRect(instance.element, event))) {
          event.preventDefault();
          event.stopPropagation();
          instance.element.classList.remove('p-winnie-container-dnd-target');
        }
      });
      Ui__default["default"].on(instance.element, Ui__default["default"].Events.DRAGOVER, function (event) {
        if (model.paletteDrag && (instance instanceof Container__default["default"] && event.target === instance.element || instance instanceof HolyGrailPane__default["default"] && event.target.parentElement === instance.element || instance instanceof TabbedPane__default["default"] && event.target.parentElement === instance.element || instance instanceof DataGrid__default["default"] && inRect(instance.element, event))) {
          event.preventDefault();
          event.stopPropagation();
          event.dropEffect = 'move';
        }
      });
      Ui__default["default"].on(instance.element, Ui__default["default"].Events.DROP, function (event) {
        if (model.paletteDrag && (instance instanceof Container__default["default"] && event.target === instance.element || instance instanceof HolyGrailPane__default["default"] && event.target.parentElement === instance.element || instance instanceof TabbedPane__default["default"] && event.target.parentElement === instance.element || instance instanceof DataGrid__default["default"] && inRect(instance.element, event))) {
          event.preventDefault();
          event.stopPropagation();
          instance.element.classList.remove('p-winnie-container-dnd-target');
          model.layout.explorer.goTo(instance['winnie.wrapper']);
          model.lastSelected = instance['winnie.wrapper'];
          var wX = Ui__default["default"].absoluteLeft(instance.element);
          var wY = Ui__default["default"].absoluteTop(instance.element);
          var left = event.clientX - wX;
          var top = event.clientY - wY;
          var applyPosition = instance instanceof Anchors__default["default"] && model.paletteDrag.item.defaultInstance instanceof Widget__default["default"];
          model.addWidget(model.paletteDrag.item, applyPosition ? Math.max(0, left) : null, applyPosition ? Math.max(0, top) : null);
          model.stickDecors();
        }
      });
    }
  }

  return instance;
}

function rename(model, created, newName) {
  if (model.widgets.has(newName)) {
    model.alert(i18n['winnie.name.used']);
    model.layout.explorer.abortEditing();
    return created.name;
  } else if (newName !== null && newName.match(/^[a-zA-Z_][a-zA-Z_0-9]*$/)) {
    var oldName = created.name;
    model.edit({
      name: "Rename widget '".concat(oldName, "' as '").concat(newName, "'"),
      redo: function redo() {
        model.widgets["delete"](oldName);
        created._name = newName;
        model.widgets.set(newName, created);
        model.layout.explorer.changed(created);
        model.layout.explorer.goTo(created, true);
      },
      undo: function undo() {
        model.widgets["delete"](newName);
        created._name = oldName;
        model.widgets.set(oldName, created);
        model.layout.explorer.changed(created);
        model.layout.explorer.goTo(created, true);
      }
    });
    model.checkEnabled();
    return newName;
  } else {
    model.alert(i18n['winnie.bad.name']);
    model.layout.explorer.abortEditing();
    return created.name;
  }
}

exports.produce = produce;
exports.produced = produced;
exports.rename = rename;
