'use strict';

var KeyCodes = require('kenga/key-codes');
var Anchors = require('kenga-containers/anchors-pane');
var locationSize = require('./location-size.js');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var KeyCodes__default = /*#__PURE__*/_interopDefaultLegacy(KeyCodes);
var Anchors__default = /*#__PURE__*/_interopDefaultLegacy(Anchors);

function surfaceKeyDown(model, event) {
  if (event.keyCode === KeyCodes__default["default"].KEY_UP || event.keyCode === KeyCodes__default["default"].KEY_DOWN || event.keyCode === KeyCodes__default["default"].KEY_LEFT || event.keyCode === KeyCodes__default["default"].KEY_RIGHT) {
    if (model.layout.explorer.selected.length > 0) {
      event.stopPropagation();
      event.preventDefault();
      model.checkEnabled();
      var moved = model.layout.explorer.selected.filter(function (item) {
        return item.delegate.attached && item.delegate.parent instanceof Anchors__default["default"];
      }).map(function (item) {
        return {
          item: item,
          startSnapshot: locationSize.sizeLocationSnapshot(item.delegate)
        };
      });

      if (moved.length > 0) {
        locationSize.proceedItemsMove(model, moved, function () {
          if (event.keyCode === KeyCodes__default["default"].KEY_UP) {
            return {
              x: 0,
              y: event.ctrlKey ? -1 : -model.settings.grid.y
            };
          } else if (event.keyCode === KeyCodes__default["default"].KEY_DOWN) {
            return {
              x: 0,
              y: event.ctrlKey ? 1 : model.settings.grid.y
            };
          } else if (event.keyCode === KeyCodes__default["default"].KEY_LEFT) {
            return {
              x: event.ctrlKey ? -1 : -model.settings.grid.x,
              y: 0
            };
          } else if (event.keyCode === KeyCodes__default["default"].KEY_RIGHT) {
            return {
              x: event.ctrlKey ? 1 : model.settings.grid.x,
              y: 0
            };
          } else {
            throw 'Unknown key for selected widgets movement.';
          }
        }(), event.ctrlKey ? false : model.settings.grid.snap);
        locationSize.endItemsMove(model, moved);
      }
    }
  }
}

function winnieKeyDown(model, event) {
  if (event.ctrlKey && event.keyCode === KeyCodes__default["default"].KEY_Z) {
    event.stopPropagation();
    event.preventDefault();
    model.checkEnabled();

    if (event.shiftKey) {
      model.redo();
    } else {
      model.undo();
    }
  } else if (event.ctrlKey && event.keyCode === KeyCodes__default["default"].KEY_Y) {
    event.stopPropagation();
    event.preventDefault();
    model.checkEnabled();
    model.redo();
  } else if (event.ctrlKey && event.keyCode === KeyCodes__default["default"].KEY_O) {
    event.stopPropagation();
    event.preventDefault();
    model.openJSON();
  } else if (event.ctrlKey && event.keyCode === KeyCodes__default["default"].KEY_S) {
    event.stopPropagation();
    event.preventDefault();
    model.save();
  } else if (event.ctrlKey && event.keyCode === KeyCodes__default["default"].KEY_E) {
    event.stopPropagation();
    event.preventDefault();
    model["export"]();
  } else if (event.keyCode === KeyCodes__default["default"].KEY_DELETE) {
    if (model.layout.explorer.selected.length > 0 && !model.layout.explorer.activeEditor && !model.layout.properties.activeEditor) {
      event.stopPropagation();
      event.preventDefault();
      model.checkEnabled();
      model.removeSelected();
    }
  } else if (event.ctrlKey && (event.keyCode === KeyCodes__default["default"].KEY_C || event.keyCode === KeyCodes__default["default"].KEY_INSERT)) {
    if (model.layout.explorer.selected.length > 0 && !model.layout.explorer.activeEditor && !model.layout.properties.activeEditor) {
      event.stopPropagation();
      event.preventDefault();
      model.checkEnabled();
      model.copy();
    }
  } else if (event.ctrlKey && event.keyCode === KeyCodes__default["default"].KEY_X || event.shiftKey && event.keyCode === KeyCodes__default["default"].KEY_DELETE) {
    if (model.layout.explorer.selected.length > 0 && !model.layout.explorer.activeEditor && !model.layout.properties.activeEditor) {
      event.stopPropagation();
      event.preventDefault();
      model.checkEnabled();
      model.cut();
    }
  }
}

var Shortcuts = {
  winnieKeyDown: winnieKeyDown,
  surfaceKeyDown: surfaceKeyDown
};

module.exports = Shortcuts;
